import sys
import itertools
import collections
import random
import math
import pygame

from . import ptext
from . import data
from . import const
from . import levels
from . import util


class GameWindow(object):
    
    def __init__(self):
        pygame.display.set_caption('Resolute')
        # self.screen = pygame.display.set_mode((const.WIDTH, const.HEIGHT), pygame.FULLSCREEN)
        self.screen = pygame.display.set_mode((const.WIDTH, const.HEIGHT))
        try:
            pygame.mixer.init()
            # pygame.mixer.music.load(data.filepath(""))
            # pygame.mixer.music.play(-1)
        except:
            pass
        self.intro()
    
    def intro(self):
        Intro(self, self.screen).loop()
        self.game()
    
    def game(self):
        for nlevel in range(len(levels.LEVELS)):
            restart = True
            while restart:
                game = Game(self, self.screen, nlevel)
                game.loop()
                restart = game.restarted


class Intro(object):

    def __init__(self, window, screen):
        self.window = window
        self.real_screen = screen
        self.screen = pygame.surface.Surface((const.REAL_WIDTH, const.REAL_HEIGHT))
        self.bg = pygame.image.load(data.filepath("intro.png")).convert()
        self.screen.blit(self.bg, (0, 0))
        self.fullscreen = False
        pygame.transform.scale(self.screen, (const.WIDTH, const.HEIGHT), self.real_screen)
        pygame.display.update()
        
    def loop(self):
        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        sys.exit()
                    elif event.key == pygame.K_f:
                        if not self.fullscreen:
                            pygame.display.set_mode((const.WIDTH, const.HEIGHT), pygame.FULLSCREEN)
                            pygame.mouse.set_visible(0)
                            self.fullscreen = True
                        else:
                            pygame.display.set_mode((const.WIDTH, const.HEIGHT))
                            self.fullscreen = False
                    else:
                        self.done = True
                        sound = pygame.mixer.Sound(data.filepath("start.ogg"))
                        sound.set_volume(const.SOUND_VOLUME)
                        sound.play()

            pygame.display.flip()


class Game(object):

    def __init__(self, window, screen, levelnum):
        self.window = window
        self.real_screen = screen
        self.screen = pygame.surface.Surface((const.REAL_WIDTH, const.REAL_HEIGHT))
        self.clock = pygame.time.Clock()

        self.paused = False

        self.level = Level(levels.LEVELS[levelnum])

        self.init_sprites()
        self.previous_res = []

        self.scaling_sprites = pygame.sprite.OrderedUpdates()

        self.grid = Grid(self.level, self.target)
        self.grid_con = pygame.sprite.RenderUpdates(self.grid)
        self.level.grid = self.grid

        instructions = pygame.sprite.Sprite()
        instructions.image = pygame.image.load(data.filepath("instructions.png")).convert_alpha()
        instructions.rect = instructions.image.get_rect()
        instructions.rect.topleft = (0, 0)
        self.instructions = pygame.sprite.RenderUpdates(instructions)
        self.show_instructions = False

        self.text_layer = TextLayer(levelnum)
        self.text_layer.draw_level_title()
        self.text_layer.draw_level_message()

        self.fullscreen = False
        self.lost = False
        self.won = False
        self.restarted = True

        self.transition = False
        self.level_str_to_transition = None
        self.do_end_decrease_resolution = False

    def loop(self):
        self.bg = pygame.image.load(data.filepath("bg.png"))
        self.screen.blit(self.bg, (0, 0))

        border_sprite = pygame.sprite.Sprite()
        border_sprite.image = pygame.image.load(data.filepath("border.png")).convert_alpha()
        border_sprite.rect = border_sprite.image.get_rect()
        border_sprite.rect.topleft = (0, 0)
        self.border = pygame.sprite.RenderUpdates(border_sprite)

        pygame.transform.scale(self.screen, (const.WIDTH, const.HEIGHT), self.real_screen)
        pygame.display.update()

        self.done = False

        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    self.on_keydown(event)
                elif event.type == pygame.KEYUP:
                    self.on_keyup(event)

            if self.update_full_screen:
                self.screen.blit(self.bg, (0, 0))
                self.update_full_screen = False

            if not self.paused:
                self.grid_con.clear(self.screen, self.bg)
                self.target_con.clear(self.screen, self.bg)
                self.boxes.clear(self.screen, self.bg)
                self.walls.clear(self.screen, self.bg)
                self.character_con.clear(self.screen, self.bg)
                self.scaling_sprites.clear(self.screen, self.bg)
                
                self.border.clear(self.screen, self.bg)
                self.text_layer.clear(self.screen, self.bg)

            if self.show_instructions:
                self.instructions.clear(self.screen, self.bg)

            dt = self.clock.tick(const.FPS) / 1000.0

            if not self.paused:
                self.grid.update(dt)
                self.character_con.update(dt)
                self.boxes.update(dt)
                self.walls.update(dt)
                self.scaling_sprites.update(dt)

            self.update(dt)

            update_rects = []
            if not self.paused:
                update_rects += self.grid_con.draw(self.screen)
                if not self.transition:
                    update_rects += self.target_con.draw(self.screen)
                    update_rects += self.boxes.draw(self.screen)
                    update_rects += self.walls.draw(self.screen)
                    update_rects += self.character_con.draw(self.screen)
                update_rects += self.scaling_sprites.draw(self.screen)
                
                update_rects += self.border.draw(self.screen)
                update_rects += self.text_layer.draw(self.screen)

            if self.show_instructions:
                update_rects += self.instructions.draw(self.screen)

            if self.do_end_decrease_resolution:
                self.end_decrease_resolution()
                self.do_end_decrease_resolution = False

            scaled_update_rects = [pygame.Rect(const.SCALE*r.left, const.SCALE*r.top,
                const.SCALE*r.width, const.SCALE*r.height) for r in update_rects]

            pygame.transform.scale(self.screen, (const.WIDTH, const.HEIGHT), self.real_screen)
            pygame.display.flip()

    def on_keydown(self, event):
        if event.key == pygame.K_ESCAPE:
            if self.paused:
                sys.exit()
            else:
                self.pause()
                if self.show_instructions:
                    self.paused = False
                    self.show_instructions = False
        elif event.key == pygame.K_p:
            self.pause()
            if self.show_instructions:
                self.paused = False
                self.show_instructions = False
        elif event.key == pygame.K_r:
            self.restarted = True
            self.done = True
        elif event.key == pygame.K_f:
            if not self.fullscreen:
                pygame.display.set_mode((const.WIDTH, const.HEIGHT), pygame.FULLSCREEN)
                pygame.mouse.set_visible(0)
                self.update_full_screen = True
                self.fullscreen = True
            else:
                pygame.display.set_mode((const.WIDTH, const.HEIGHT))
                self.update_full_screen = True
                self.fullscreen = False
        elif event.key == pygame.K_SPACE:
            if self.won:
                self.next_level()
            if self.show_instructions:
                self.paused = False
                self.show_instructions = False
        elif event.key == pygame.K_i:
            if self.show_instructions:
                self.paused = False
                self.show_instructions = False
            else:
                self.pause()
                self.show_instructions = True
        if not self.won and not self.transition and not self.paused:
            if event.key == pygame.K_UP or event.key == pygame.K_w:
                self.character.move_up()
            elif event.key == pygame.K_DOWN or event.key == pygame.K_s:
                self.character.move_down()
            elif event.key == pygame.K_LEFT or event.key == pygame.K_a:
                self.character.move_left()
            elif event.key == pygame.K_RIGHT or event.key == pygame.K_d:
                self.character.move_right()

    def on_keyup(self, event):
        if not self.won and not self.transition and not self.paused:
            if event.key == pygame.K_UP or event.key == pygame.K_w:
                self.character.stop_up()
            elif event.key == pygame.K_DOWN or event.key == pygame.K_s:
                self.character.stop_down()
            elif event.key == pygame.K_LEFT or event.key == pygame.K_a:
                self.character.stop_left()
            elif event.key == pygame.K_RIGHT or event.key == pygame.K_d:
                self.character.stop_right()
            elif event.key == pygame.K_q:
                self.start_decrease_resolution(self.level.decrease_resolution())
            elif event.key == pygame.K_e:
                self.increase_resolution()
        if event.key == pygame.K_r:
            self.restart()
        elif event.key == pygame.K_n:
            self.win()
            self.next_level()

    def pause(self):
        self.paused = not self.paused

    def restart(self):
        self.done = True
        self.restarted = True

    def win(self):
        if not self.won:
            self.won = True
            self.text_layer.draw_win_text()
            self.character.set_image_win()
            sound = pygame.mixer.Sound(data.filepath("win.ogg"))
            sound.set_volume(const.SOUND_VOLUME)
            sound.play()

    def next_level(self):
        self.done = True
        self.restarted = False

    def update(self, dt):
        if self.level.end == self.character.tilepos:
            self.win()

    def start_decrease_resolution(self, level_str):
        if not level_str:
            sound = pygame.mixer.Sound(data.filepath("deny.ogg"))
            sound.set_volume(const.SOUND_VOLUME)
            sound.play()
            return
        self.level.on_hold = True
        self.previous_res.append((self.level, self.character, self.character_con, self.target, self.target_con,
            self.walls, self.boxes))
        self.level_str_to_transition = level_str
        self.create_scaling_sprites()

        sound = pygame.mixer.Sound(data.filepath("scale_up.ogg"))
        sound.set_volume(const.SOUND_VOLUME)
        sound.play()

    def end_decrease_resolution(self):
        assert(self.level_str_to_transition)
        self.scaling_sprites = pygame.sprite.OrderedUpdates()
        self.level = Level(self.level_str_to_transition)
        self.level.grid = self.grid
        self.init_sprites()
        previous_level = self.previous_res[-1][0]
        self.level.set_objs_children(previous_level)
        self.level_str_to_transition = None

    def increase_resolution(self):
        if len(self.previous_res) == 0 or self.previous_res[-1][0].check_conflict_increase_resolution():
            sound = pygame.mixer.Sound(data.filepath("deny.ogg"))
            sound.set_volume(const.SOUND_VOLUME)
            sound.play()
            return
        self.level, self.character, self.character_con, self.target, self.target_con, self.walls, self.boxes = self.previous_res.pop()
        self.level.on_hold = False
        self.level.update_pos_by_parents()
        self.level.unlink_parents()
        self.update_full_screen = True

        sound = pygame.mixer.Sound(data.filepath("scale_down.ogg"))
        sound.set_volume(const.SOUND_VOLUME)
        sound.play()

    def done_scaling(self):
        if self.transition:
            self.transition = False
            self.do_end_decrease_resolution = True

    def create_scaling_sprites(self):
        self.transition = True
        self.scaling_sprites = pygame.sprite.OrderedUpdates()
        level_to_transition = self.level_str_to_transition.split('\n')
        order = []
        for obj in self.level.objs.values():
            new_tilepos = (int(obj.tilepos[0] / 2), int(obj.tilepos[1] / 2))
            if level_to_transition[new_tilepos[1]][new_tilepos[0]] == obj.symbol:
                order.append(obj)
        symbol_priority = ['P', 'X', '#']
        for symbol in symbol_priority:
            for obj in self.level.objs.values():
                if obj.symbol == symbol and obj not in order:
                    order.append(obj)
        for obj in self.level.objs.values():
            new_tilepos = (int(obj.tilepos[0] / 2), int(obj.tilepos[1] / 2))
            if level_to_transition[new_tilepos[1]][new_tilepos[0]] == '.':
                order.remove(obj)
        order.reverse()
        for sprite in order:
            scaling_sprite = TransitionScalingSprite(sprite, self.done_scaling)
            scaling_sprite.start_scale(True)
            self.scaling_sprites.add(scaling_sprite)

    def init_sprites(self):
        self.character = Character(self.level)
        self.character_con = pygame.sprite.RenderUpdates(self.character)
        self.target = None
        self.target_con = pygame.sprite.RenderUpdates()
        if self.level.end:
            self.target = Target(self.level)
            self.target_con.add(self.target)
        self.walls = Walls(self.level)
        self.boxes = Boxes(self.level)
        self.update_full_screen = True


class Level():

    def __init__(self, level):
        self.parse_level(level)
        self.level = self.starting_level
        self.resolution = self.starting_resolution
        self.start = self.find('P')[0]
        self.end = self.find('>')
        self.end = self.end[0] if len(self.end) > 0 else None
        self.tile_size = const.TILE_SIZE / (2 ** self.resolution)
        self.objs = {}
        self.width = len(self.level[0])
        self.height = len(self.level)
        self.on_hold = False

    def parse_level(self, level):
        self.starting_level = [list(line) for line in level.split('\n') if line != ""]
        self.starting_resolution = round(math.log(len(self.starting_level[0]) / 3, 2))

    def find(self, obj):
        return self.find_obj(obj, self.level)

    def find_obj(self, obj, level):
        positions = []
        for i in range(len(level)):
            for j in range(len(level[i])):
                if level[i][j] == obj:
                    positions.append((j, i))
        return positions

    def check_move(self, direction, tilepos):
        if tilepos not in self.objs:
            return True
        obj = self.objs[tilepos]
        if isinstance(obj, MovableSprite):
            if isinstance(obj, Character):
                return True  # allow self loops
            if obj.pushable:
                return self.check_move(direction, util.get_target_tilepos(self, tilepos, direction)[0])
            else:
                return False
        else:
            return False

    def add_obj(self, tilepos, obj, recompute=True):
        self.objs[tilepos] = obj
        if recompute:
            self.recompute_level()

    def recompute_level(self):
        self.level = [['.' for i in range(self.width)] for j in range(self.height)]
        for tilepos, obj in self.objs.items():
            self.level[tilepos[1]][tilepos[0]] = obj.symbol
        if self.end:
            self.level[self.end[1]][self.end[0]] = '>'
        self.grid.draw_grid()

    def get_block(self, i, j):
        return "".join(["".join([self.level[jj][ii] for ii in range(i*2, i*2 + 2)]) for jj in range(j*2, j*2 + 2)])

    def decrease_resolution_block(self, block):
        block = block.replace('P', '.')
        block = block.replace('>', '.')
        counter = collections.defaultdict(int)
        for c in block:
            counter[c] += 1
        max_val = max(counter.values())
        max_symbols = [k for k, v in counter.items() if v == max_val]
        precedence = ['X', '#', '.']
        for symbol in precedence:
            if symbol in max_symbols:
                return symbol
        assert(False)
        return None

    def decrease_resolution(self):
        if self.resolution == 0:
            return None
        new_width = int(self.width / 2)
        new_height = int(self.height / 2)
        new_level = [['.' for i in range(new_width)] for j in range(new_height)]
        for i in range(new_width):
            for j in range(new_height):
                new_level[j][i] = self.decrease_resolution_block(self.get_block(i, j))
        player_pos = self.find('P')[0]
        player_pos = (int(player_pos[0] / 2), int(player_pos[1] / 2))
        if new_level[player_pos[1]][player_pos[0]] in ['#', 'X']:
            return None
        new_level[player_pos[1]][player_pos[0]] = 'P'

        new_level_str = "\n".join(["".join(new_level[i]) for i in range(len(new_level))])
        return new_level_str

    def set_objs_children(self, res_level):
        for tilepos, obj in self.objs.items():
            children = []
            for offset in [(0,0), (0,1), (1,0), (1,1)]:
                res_tilepos = (2 * tilepos[0] + offset[0], 2 * tilepos[1] + offset[1])
                if res_tilepos in res_level.objs:
                    children.append(res_level.objs[res_tilepos])
            obj.set_children(children)

    def compute_objs_by_tilepos_parents(self):
        objs_by_tilepos = collections.defaultdict(list)
        for obj in self.objs.values():
            if obj.parent:
                new_tilepos = (2 * obj.parent.tilepos[0] + obj.tilepos[0] % 2,
                    2 * obj.parent.tilepos[1] + obj.tilepos[1] % 2)
                objs_by_tilepos[new_tilepos].append(obj)
            else:
                objs_by_tilepos[obj.tilepos].append(obj)
        return objs_by_tilepos

    def check_conflict_increase_resolution(self):
        objs_by_tilepos = self.compute_objs_by_tilepos_parents()
        for tilepos, obj_list in objs_by_tilepos.items():
            if len(obj_list) > 1:
                return True  # conflict
            elif len(obj_list) == 1 and tilepos == self.end:
                return True
        return False

    def update_pos_by_parents(self):
        all_objs = list(self.objs.values())
        for obj in all_objs:
            if obj.parent:
                assert(self.objs[obj.tilepos] == obj)
                self.objs.pop(obj.tilepos)
        for obj in all_objs:
            if obj.parent:
                new_tilepos = (2 * obj.parent.tilepos[0] + obj.tilepos[0] % 2,
                    2 * obj.parent.tilepos[1] + obj.tilepos[1] % 2)
                obj.update_tilepos(new_tilepos)
                self.add_obj(obj.tilepos, obj)

    def unlink_parents(self):
        for obj in self.objs.values():
            obj.parent = None


class BaseSprite(pygame.sprite.Sprite):

    def __init__(self, level, tilepos):
        pygame.sprite.Sprite.__init__(self)
        self.level = level
        self.update_tilepos(tilepos)

        self.parent = None
        self.children = {}
        self.scaling = False

        self.vel = [0, 0]

    def update_tilepos(self, tilepos):
        self.tilepos = tuple(tilepos)
        self.pos = list(util.get_xy_at_tile(self.level, *self.tilepos))

    def set_children(self, children):
        self.children = {}
        for child in children:
            child.parent = self
            self.children[(child.tilepos[0] % 2, child.tilepos[1] % 2)] = child

    def update(self, dt):
        if self.parent:
            self.pos[0] = self.parent.pos[0] + (self.tilepos[0] % 2) * self.level.tile_size
            self.pos[1] = self.parent.pos[1] + (self.tilepos[1] % 2) * self.level.tile_size
            self.rect.topleft = self.pos


class TransitionScalingSprite(BaseSprite):

    def __init__(self, sprite, done_scaling_func):
        BaseSprite.__init__(self, sprite.level, sprite.tilepos)
        self.image = sprite.image.copy()
        self.rect = self.image.get_rect()
        self.done_scaling_func = done_scaling_func

        self.scaling = False
        self.scaling_vel = 0
        self.scaling_size = 0
        self.scaling_target = 0
        self.scaling_image_before = None
        self.scaling_top = True
        self.scaling_left = True

    def start_scale(self, up):
        self.scaling_vel = 100 if up else -100
        self.scaling_size = self.image.get_width()
        self.scaling_target = 2 * self.scaling_size if up else self.scaling_size / 2
        self.scaling = True
        self.scaling_image_before = self.image
        self.scaling_top = True if (self.tilepos[1] % 2 == 0) else False
        self.scaling_left = True if (self.tilepos[0] % 2 == 0) else False

    def stop_scale(self):
        self.scaling_vel = 0
        self.scaling = False
        self.done_scaling_func()

    def scaling_set_pos(self, scaling_size):
        if self.scaling_top:
            if not self.scaling_left:
                pos = util.get_xy_at_tile(self.level, *self.tilepos)
                self.pos = [pos[0] + self.level.tile_size - scaling_size, pos[1]]
        else:
            if self.scaling_left:
                pos = util.get_xy_at_tile(self.level, *self.tilepos)
                self.pos = [pos[0], pos[1] + self.level.tile_size - scaling_size]
            else:
                pos = util.get_xy_at_tile(self.level, *self.tilepos)
                self.pos = [pos[0] + self.level.tile_size - scaling_size, pos[1] + self.level.tile_size - scaling_size]

    def update(self, dt):
        if self.scaling:
            self.scaling_size += self.scaling_vel * dt
            if ((self.scaling_vel > 0 and self.scaling_size >= self.scaling_target)
                or (self.scaling_vel < 0 and self.scaling_size <= self.scaling_target)):
                self.scaling_size = self.scaling_target
                self.stop_scale()
            int_size = int(self.scaling_size)
            self.image = pygame.transform.scale(self.scaling_image_before, (int_size, int_size))
            self.scaling_set_pos(int_size)
            self.rect.topleft = self.pos


class MovableSprite(BaseSprite):

    def __init__(self, level, tilepos):
        BaseSprite.__init__(self, level, tilepos)
        self.last_pos = list(self.pos)
        self.vel = [0, 0]

        self.speed = const.SPEED

        self.moving = False
        self.pressed_dir = {(-1, 0) : False, (1, 0) : False, (0, -1) : False, (0, 1) : False}

        self.pushable = False
        self.looping = False

    def update(self, dt):
        BaseSprite.update(self, dt)
        if not self.parent:
            self.update_move()
            self.last_pos[0] = self.pos[0]
            self.last_pos[1] = self.pos[1]
            self.pos[0] += self.vel[0] * dt
            self.pos[1] += self.vel[1] * dt
            self.rect.topleft = self.pos

    def move_left(self):  self.pressed_dir[(-1, 0)] = True
    def stop_left(self):  self.pressed_dir[(-1, 0)] = False
    def move_right(self): self.pressed_dir[(+1, 0)] = True
    def stop_right(self): self.pressed_dir[(+1, 0)] = False
    def move_up(self):    self.pressed_dir[(0, -1)] = True
    def stop_up(self):    self.pressed_dir[(0, -1)] = False
    def move_down(self):  self.pressed_dir[(0, +1)] = True
    def stop_down(self):  self.pressed_dir[(0, +1)] = False

    def try_move_once(self, dir):
        target_tilepos, looped_tilepos = util.get_target_tilepos(self.level, self.tilepos, dir) 

        if not self.level.check_move(dir, target_tilepos):
            return False

        self.target_tilepos = target_tilepos
        self.target = util.get_xy_at_tile(self.level, *target_tilepos)
        self.vel[0] = dir[0] * self.speed
        self.vel[1] = dir[1] * self.speed
        self.moving = True

        assert self.level.objs[self.tilepos] == self

        if target_tilepos in self.level.objs:
            pushed_obj = self.level.objs[target_tilepos]
            if isinstance(pushed_obj, MovableSprite) and not isinstance(pushed_obj, Character):
                if pushed_obj.moving:
                    pushed_obj.finish_move()
                pushed_obj.try_move_once(dir)

        if self.level.objs[self.tilepos] == self:  # do not pop if looping
            self.level.objs.pop(self.tilepos)
        self.level.add_obj(target_tilepos, self)

        if looped_tilepos != self.tilepos:
            self.tilepos = looped_tilepos
            self.pos = list(util.get_xy_at_tile(self.level, *self.tilepos))
            self.looping = True

        return True

    def finish_move(self):
        self.pos = list(self.target)
        self.tilepos = tuple(self.target_tilepos)
        self.vel = [0, 0]
        self.moving = False
        self.looping = False

    def update_move(self):
        if self.moving:
            if ((self.vel[0] > 0 and self.pos[0] >= self.target[0])
                or (self.vel[0] < 0 and self.pos[0] <= self.target[0])
                or (self.vel[1] > 0 and self.pos[1] >= self.target[1])
                or (self.vel[1] < 0 and self.pos[1] <= self.target[1])):
                self.finish_move()
        if not self.moving:
            for d, pressed in self.pressed_dir.items():
                if pressed:
                    if self.try_move_once(d):
                        break


class Character(MovableSprite):

    def __init__(self, level):
        MovableSprite.__init__(self, level, level.start)

        self.image = pygame.image.load(data.filepath("character.png")).convert_alpha()
        scale = 2 ** (2 - self.level.resolution)
        self.image = pygame.transform.scale(self.image, (int(self.image.get_width() * scale), int(self.image.get_height() * scale)))
        self.rect = self.image.get_rect()

        self.image_win = pygame.image.load(data.filepath("character_win.png")).convert_alpha()
        scale = 2 ** (2 - self.level.resolution)
        self.image_win = pygame.transform.scale(self.image_win, (int(self.image_win.get_width() * scale), int(self.image_win.get_height() * scale)))

        self.symbol = 'P'
        self.level.add_obj(self.level.start, self, recompute=False)
        self.pushable = True

    def set_image_win(self):
        self.image = self.image_win


class Wall(BaseSprite):

    def __init__(self, level, tilepos):
        BaseSprite.__init__(self, level, tilepos)
        self.image = pygame.image.load(data.filepath("wall.png")).convert_alpha()
        scale = 2 ** (2 - self.level.resolution)
        self.image = pygame.transform.scale(self.image, (int(self.image.get_width() * scale), int(self.image.get_height() * scale)))
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.symbol = '#'
        self.level.add_obj(tilepos, self, recompute=False)

    def update(self, dt):
        BaseSprite.update(self, dt)
        self.rect.topleft = self.pos


class Walls(pygame.sprite.RenderUpdates):

    def __init__(self, level):
        pygame.sprite.RenderUpdates.__init__(self)
        self.level = level
        for wall_pos in self.level.find("#"):
            self.add(Wall(self.level, wall_pos))


class Box(MovableSprite):

    def __init__(self, level, tilepos):
        MovableSprite.__init__(self, level, tilepos)
        self.image = pygame.image.load(data.filepath("box.png")).convert_alpha()
        scale = 2 ** (2 - self.level.resolution)
        self.image = pygame.transform.scale(self.image, (int(self.image.get_width() * scale), int(self.image.get_height() * scale)))
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.symbol = 'X'
        self.level.add_obj(tilepos, self, recompute=False)
        self.pushable = True


class Boxes(pygame.sprite.RenderUpdates):

    def __init__(self, level):
        pygame.sprite.RenderUpdates.__init__(self)
        self.level = level
        for box_pos in self.level.find("X"):
            self.add(Box(self.level, box_pos))


class Target(pygame.sprite.Sprite):

    def __init__(self, level):
        pygame.sprite.Sprite.__init__(self)
        self.level = level
        self.pos = list(util.get_xy_at_tile(self.level, *self.level.end))
        self.image = pygame.image.load(data.filepath("target.png")).convert_alpha()
        scale = 2 ** (2 - self.level.resolution)
        self.image = pygame.transform.scale(self.image, (int(self.image.get_width() * scale), int(self.image.get_height() * scale)))
        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
        self.symbol = '>'


class Grid(pygame.sprite.Sprite):

    def __init__(self, level, target):
        pygame.sprite.Sprite.__init__(self)
        self.level = level
        self.target = target
        if self.level.resolution > 0:
            self.draw_grid()
        else:
            self.image = pygame.surface.Surface((0, 0))
            self.rect = self.image.get_rect()
            self.rect.topleft = (0, 0)

    def draw_corner(self, i, j, color, color2=None):
        if color2 == color:
            color2 = (max(0, color2[0] - 20), max(0, color2[1] - 20), max(0, color2[2] - 20))
        if j % 2 == 0:
            pygame.draw.line(self.image, color, 
                (const.MAP_OFFSET_X + i * self.level.tile_size, const.MAP_OFFSET_Y + j * self.level.tile_size),
                (const.MAP_OFFSET_X + (i + 1) * self.level.tile_size - 1, const.MAP_OFFSET_Y + j * self.level.tile_size), 2)
            if color2:
                pygame.draw.line(self.image, color2, 
                    (const.MAP_OFFSET_X + i * self.level.tile_size, const.MAP_OFFSET_Y + j * self.level.tile_size),
                    (int(const.MAP_OFFSET_X + (i + 0.25) * self.level.tile_size - 1), const.MAP_OFFSET_Y + j * self.level.tile_size), 2)
                pygame.draw.line(self.image, color2, 
                    (int(const.MAP_OFFSET_X + (i + 0.5) * self.level.tile_size - 1), const.MAP_OFFSET_Y + j * self.level.tile_size),
                    (int(const.MAP_OFFSET_X + (i + 0.75) * self.level.tile_size - 1), const.MAP_OFFSET_Y + j * self.level.tile_size), 2)
        if j % 2 == 1:
            pygame.draw.line(self.image, color, 
                (const.MAP_OFFSET_X + i * self.level.tile_size, const.MAP_OFFSET_Y + (j + 1) * self.level.tile_size - 2),
                (const.MAP_OFFSET_X + (i + 1) * self.level.tile_size - 1, const.MAP_OFFSET_Y + (j + 1) * self.level.tile_size - 2), 2)
            if color2:
                pygame.draw.line(self.image, color2, 
                    (const.MAP_OFFSET_X + i * self.level.tile_size, const.MAP_OFFSET_Y + (j + 1) * self.level.tile_size - 2),
                    (int(const.MAP_OFFSET_X + (i + 0.25) * self.level.tile_size - 1), const.MAP_OFFSET_Y + (j + 1) * self.level.tile_size - 2), 2)
                pygame.draw.line(self.image, color2, 
                    (int(const.MAP_OFFSET_X + (i + 0.5) * self.level.tile_size - 1), const.MAP_OFFSET_Y + (j + 1) * self.level.tile_size - 2),
                    (int(const.MAP_OFFSET_X + (i + 0.75) * self.level.tile_size - 1), const.MAP_OFFSET_Y + (j + 1) * self.level.tile_size - 2), 2)
        if i % 2 == 0:
            pygame.draw.line(self.image, color, 
                (const.MAP_OFFSET_X + i * self.level.tile_size, const.MAP_OFFSET_Y + j * self.level.tile_size),
                (const.MAP_OFFSET_X + i * self.level.tile_size, const.MAP_OFFSET_Y + (j + 1) * self.level.tile_size - 1), 2)
            if color2:
                pygame.draw.line(self.image, color2, 
                    (const.MAP_OFFSET_X + i * self.level.tile_size, const.MAP_OFFSET_Y + j * self.level.tile_size),
                    (const.MAP_OFFSET_X + i * self.level.tile_size, int(const.MAP_OFFSET_Y + (j + 0.25) * self.level.tile_size - 1)), 2)
                pygame.draw.line(self.image, color2, 
                    (const.MAP_OFFSET_X + i * self.level.tile_size, int(const.MAP_OFFSET_Y + (j + 0.5) * self.level.tile_size - 1)),
                    (const.MAP_OFFSET_X + i * self.level.tile_size, int(const.MAP_OFFSET_Y + (j + 0.75) * self.level.tile_size - 1)), 2)
        if i % 2 == 1:
            pygame.draw.line(self.image, color, 
                (const.MAP_OFFSET_X + (i + 1) * self.level.tile_size - 2, const.MAP_OFFSET_Y + j * self.level.tile_size),
                (const.MAP_OFFSET_X + (i + 1) * self.level.tile_size - 2, const.MAP_OFFSET_Y + (j + 1) * self.level.tile_size - 1), 2)
            if color2:
                pygame.draw.line(self.image, color2, 
                    (const.MAP_OFFSET_X + (i + 1) * self.level.tile_size - 2, const.MAP_OFFSET_Y + j * self.level.tile_size),
                    (const.MAP_OFFSET_X + (i + 1) * self.level.tile_size - 2, int(const.MAP_OFFSET_Y + (j + 0.25) * self.level.tile_size - 1)), 2)
                pygame.draw.line(self.image, color2, 
                    (const.MAP_OFFSET_X + (i + 1) * self.level.tile_size - 2, int(const.MAP_OFFSET_Y + (j + 0.5) * self.level.tile_size - 1)),
                    (const.MAP_OFFSET_X + (i + 1) * self.level.tile_size - 2, int(const.MAP_OFFSET_Y + (j + 0.75) * self.level.tile_size - 1)), 2)

    def draw_base_grid(self):
        for i in range(self.level.width):
            for j in range(self.level.height):
                self.draw_corner(i, j, (64, 64, 64))

    def draw_grid(self):
        if self.level.resolution == 0:
            return  # no grid for tutorial levels

        if self.level.on_hold:
            objs = self.level.compute_objs_by_tilepos_parents()  # Very hacky. No time. :(
        else:
            objs = {k: [v] for k, v in self.level.objs.items()}
        if self.level.end not in objs:
            objs[self.level.end] = []
        objs[self.level.end].append(self.target)
        self.image = pygame.surface.Surface((const.REAL_WIDTH, const.REAL_HEIGHT))

        color_by_symbol = {'P': (45, 126, 175), '#': (185, 55, 74), 'X': (170, 116, 50), '>': (107, 190, 98)}

        self.draw_base_grid()
        for tilepos, obj in objs.items():
            i, j = tilepos
            color = (64, 64, 64)
            if len(obj) == 1:
                self.draw_corner(i, j, color_by_symbol[obj[0].symbol])
            else:
                self.draw_corner(i, j, color_by_symbol[obj[0].symbol], color_by_symbol[obj[1].symbol])
        self.rect = self.image.get_rect()
        self.rect.topleft = (0, 0)

    def update(self, dt):
        self.draw_grid()


class TextLayer(pygame.sprite.RenderUpdates):

    def __init__(self, levelnum):
        pygame.sprite.RenderUpdates.__init__(self)
        self.levelnum = levelnum
        self.level_message = None

    def draw_level_title(self):
        sprite = pygame.sprite.Sprite()
        sprite.image, pos = ptext.draw('Level ' + str(self.levelnum+1) + '/' + str(len(levels.LEVELS)), (5, 5), color=(255,255,255),
            fontname=const.FONT_MAIN, fontsize=8)
        sprite.rect = sprite.image.get_rect()
        sprite.rect.topleft = pos
        self.add(sprite)

    def draw_win_text(self):
        if self.level_message:
            self.remove(self.level_message)
        sprite = pygame.sprite.Sprite()
        if self.levelnum == len(levels.LEVELS) - 1:
            text = "Excellent! You beat the game!"
        else:
            text = random.choice(["Nice!", "Great!", "You got it!"])
            text += " Press space for next level."
        sprite.image, pos = ptext.draw(text, (10, const.REAL_HEIGHT - 15), color=(255,255,255),
            fontname=const.FONT_MAIN, fontsize=8)
        sprite.rect = sprite.image.get_rect()
        sprite.rect.topleft = pos
        self.add(sprite)

    def draw_level_message(self):
        if self.levelnum == 0:
            text = "WASD/Arrows to move. Push box & get to flag."
        elif self.levelnum == 1:
            text = "You (and blocks) can loop around."
        elif self.levelnum == 2:
            text = "Blocked off! Press I for instructions."
        elif self.levelnum == 3:
            text = "Don't forget you can loop around!"
        else:
            return
        sprite = pygame.sprite.Sprite()
        sprite.image, pos = ptext.draw(text, (10, const.REAL_HEIGHT - 15), color=(255,255,255),
            fontname=const.FONT_MAIN, fontsize=8)
        sprite.rect = sprite.image.get_rect()
        sprite.rect.topleft = pos
        self.level_message = sprite
        self.add(sprite)


if __name__ == '__main__':
    GameWindow()
